"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.sendTransactionBatch = sendTransactionBatch;
exports.sendTransaction = sendTransaction;
const viem_1 = require("viem");
const AccountFactory_js_1 = require("../abis/AccountFactory.js");
const constants_js_1 = require("../constants.js");
const utils_js_1 = require("../utils.js");
const sendPrivyTransaction_js_1 = require("./sendPrivyTransaction.js");
const sendTransactionInternal_js_1 = require("./sendTransactionInternal.js");
async function sendTransactionBatch(client, signerClient, publicClient, parameters, isPrivyCrossApp = false) {
    if (parameters.calls.length === 0) {
        throw new Error('No calls provided');
    }
    if (isPrivyCrossApp) {
        return await (0, sendPrivyTransaction_js_1.sendPrivyTransaction)(client, parameters);
    }
    const calls = parameters.calls.map((tx) => {
        if (!tx.to)
            throw new Error('Transaction target (to) is required');
        return {
            target: tx.to,
            allowFailure: false,
            value: BigInt(tx.value ?? 0),
            callData: tx.data ?? '0x',
        };
    });
    const batchCallData = (0, viem_1.encodeFunctionData)({
        abi: [
            {
                name: 'batchCall',
                type: 'function',
                inputs: [
                    {
                        type: 'tuple[]',
                        name: 'calls',
                        components: [
                            { name: 'target', type: 'address' },
                            { name: 'allowFailure', type: 'bool' },
                            { name: 'value', type: 'uint256' },
                            { name: 'callData', type: 'bytes' },
                        ],
                    },
                ],
                outputs: [],
            },
        ],
        args: [calls],
    });
    const totalValue = calls.reduce((sum, call) => sum + BigInt(call.value), BigInt(0));
    let batchTransaction;
    const isDeployed = await (0, utils_js_1.isSmartAccountDeployed)(publicClient, client.account.address);
    if (!isDeployed) {
        const initialCall = {
            target: constants_js_1.BATCH_CALLER_ADDRESS,
            allowFailure: false,
            value: totalValue,
            callData: batchCallData,
        };
        const initializerCallData = (0, utils_js_1.getInitializerCalldata)(signerClient.account.address, constants_js_1.VALIDATOR_ADDRESS, initialCall);
        const addressBytes = (0, viem_1.toBytes)(signerClient.account.address);
        const salt = (0, viem_1.keccak256)(addressBytes);
        const deploymentCalldata = (0, viem_1.encodeFunctionData)({
            abi: AccountFactory_js_1.default,
            functionName: 'deployAccount',
            args: [salt, initializerCallData],
        });
        batchTransaction = {
            to: constants_js_1.SMART_ACCOUNT_FACTORY_ADDRESS,
            data: deploymentCalldata,
            value: totalValue,
            paymaster: parameters.paymaster,
            paymasterInput: parameters.paymasterInput,
            type: 'eip712',
        };
    }
    else {
        batchTransaction = {
            to: constants_js_1.BATCH_CALLER_ADDRESS,
            data: batchCallData,
            value: totalValue,
            paymaster: parameters.paymaster,
            paymasterInput: parameters.paymasterInput,
            type: 'eip712',
        };
    }
    return (0, sendTransactionInternal_js_1.sendTransactionInternal)(client, signerClient, publicClient, batchTransaction, !isDeployed);
}
async function sendTransaction(client, signerClient, publicClient, parameters, isPrivyCrossApp = false) {
    if (isPrivyCrossApp)
        return await (0, sendPrivyTransaction_js_1.sendPrivyTransaction)(client, parameters);
    const isDeployed = await (0, utils_js_1.isSmartAccountDeployed)(publicClient, client.account.address);
    if (!isDeployed) {
        const initialCall = {
            target: parameters.to,
            allowFailure: false,
            value: parameters.value ?? 0,
            callData: parameters.data ?? '0x',
        };
        const initializerCallData = (0, utils_js_1.getInitializerCalldata)(signerClient.account.address, constants_js_1.VALIDATOR_ADDRESS, initialCall);
        const addressBytes = (0, viem_1.toBytes)(signerClient.account.address);
        const salt = (0, viem_1.keccak256)(addressBytes);
        const deploymentCalldata = (0, viem_1.encodeFunctionData)({
            abi: AccountFactory_js_1.default,
            functionName: 'deployAccount',
            args: [salt, initializerCallData],
        });
        parameters.to = constants_js_1.SMART_ACCOUNT_FACTORY_ADDRESS;
        parameters.data = deploymentCalldata;
        parameters.value = 0n;
    }
    return (0, sendTransactionInternal_js_1.sendTransactionInternal)(client, signerClient, publicClient, parameters, !isDeployed);
}
//# sourceMappingURL=sendTransaction.js.map